# Copyright (c) 2023 David Boetius
# Licensed under the MIT license
from tempfile import TemporaryDirectory

import pytest

from fairnessdatasets import Adult, SouthGerman, Default, LawSchool


@pytest.fixture(scope="session")
def adult_path():
    with TemporaryDirectory() as tmp_dir:
        Adult(tmp_dir, train=False, download=True)
        yield tmp_dir


@pytest.fixture(scope="session")
def adult_raw_path(adult_path):
    Adult(adult_path, train=False, raw=True, download=True)
    yield adult_path


@pytest.fixture(scope="session")
def default_path():
    with TemporaryDirectory() as tmp_dir:
        Default(tmp_dir, download=True)
        yield tmp_dir


@pytest.fixture(scope="session")
def default_raw_path(default_path):
    Default(default_path, raw=True, download=True)
    yield default_path


@pytest.fixture(scope="session")
def law_school_path():
    with TemporaryDirectory() as tmp_dir:
        LawSchool(tmp_dir, download=True)
        yield tmp_dir


@pytest.fixture(scope="session")
def law_school_raw_path(default_path):
    LawSchool(default_path, raw=True, download=True)
    yield default_path


@pytest.fixture(scope="session")
def south_german_path():
    with TemporaryDirectory() as tmp_dir:
        SouthGerman(tmp_dir, download=True)
        yield tmp_dir


@pytest.fixture(scope="session")
def south_german_raw_path(south_german_path):
    SouthGerman(south_german_path, raw=True, download=True)
    yield south_german_path



# Copyright (c) 2024 David Boetius
# Licensed under the MIT license
import itertools
from typing import Callable, Dict, Optional, Tuple, Union

import os

import pandas

from .base import DefaultPreprocessing


class Adult(DefaultPreprocessing):
    """
    The `Adult <https://archive.ics.uci.edu/dataset/2/adult>`_ dataset.

    The dataset is preprocessed by:
     - removing rows (samples) with missing values
     - one-hot encoding all categorical attributes
     - applying z-score normalization to all continuous variables
    The last two preprocessing steps are optional and can be turned off by
    passing :code:`raw=True` to the initializer.

    Class Attributes:
    - `dataset_url`: The URL the Adult dataset is downloaded from.
    - `files_to_download`: The files that are downloaded from `dataset_url`.
    - `checksums`: The checksums of the files in `files`.
    - `variables`: The variables of the Adult dataset,
       together with the values they may take on.
       For integer variables, such as `age`, the value is :code:`None`.
       That is, :code:`Adult.variables["age"] = None`.
       For categorical variables, like `sex`, the value is a tuple of strings.
       For example, :code:`Adult.variables["sex"] = ("female", "male")`.
       Values that do not appear in the dataset after preprocessing are not
       included.
       This only affects the values of `workclass`, where `workclass=Never-worked`
       is dropped, as it does not appear in the dataset after dropping rows with
       missing values.

    Attributes:
    - `columns`: Column labels for the tensors in this dataset
      (after one-hot encoding, if applied).
       This is :code:`Adult.columns = ("age", "workclass=Private",
       "workclass=Self-emp-not-inc", ...)`.
    - `files_dir`: Where the data files are stored or downloaded to.
      Value: Root path (user specified) / `type(self).__name__)`
    - `data`: The dataset features (x values)
    - `targets` The dataset targets (y values)
    """

    dataset_url = "https://archive.ics.uci.edu/static/public/2/adult.zip"
    files_to_download = {"test": "adult.test", "train": "adult.data"}
    checksums = {
        "adult.test": "a2a9044bc167a35b2361efbabec64e89d69ce82d9790d2980119aac5fd7e9c05",
        "adult.data": "5b00264637dbfec36bdeaab5676b0b309ff9eb788d63554ca0a249491c86603d",
    }
    variables = {
        "age": None,  # continuous variables marked with None
        "workclass": (
            "Private",
            "Self-emp-not-inc",
            "Self-emp-inc",
            "Federal-gov",
            "Local-gov",
            "State-gov",
            "Without-pay",
            # "Never-worked",  # does not appear in dataset
        ),
        "fnlwgt": None,
        "education": (
            "Bachelors",
            "Some-college",
            "11th",
            "HS-grad",
            "Prof-school",
            "Assoc-acdm",
            "Assoc-voc",
            "9th",
            "7th-8th",
            "12th",
            "Masters",
            "1st-4th",
            "10th",
            "Doctorate",
            "5th-6th",
            "Preschool",
        ),
        "education-num": None,
        "marital-status": (
            "Married-civ-spouse",
            "Divorced",
            "Never-married",
            "Separated",
            "Widowed",
            "Married-spouse-absent",
            "Married-AF-spouse",
        ),
        "occupation": (
            "Tech-support",
            "Craft-repair",
            "Other-service",
            "Sales",
            "Exec-managerial",
            "Prof-specialty",
            "Handlers-cleaners",
            "Machine-op-inspct",
            "Adm-clerical",
            "Farming-fishing",
            "Transport-moving",
            "Priv-house-serv",
            "Protective-serv",
            "Armed-Forces",
        ),
        "relationship": (
            "Wife",
            "Own-child",
            "Husband",
            "Not-in-family",
            "Other-relative",
            "Unmarried",
        ),
        "race": ("White", "Asian-Pac-Islander", "Amer-Indian-Eskimo", "Other", "Black"),
        "sex": ("Female", "Male"),
        "capital-gain": None,
        "capital-loss": None,
        "hours-per-week": None,
        "native-country": (
            "United-States",
            "Cambodia",
            "England",
            "Puerto-Rico",
            "Canada",
            "Germany",
            "Outlying-US(Guam-USVI-etc)",
            "India",
            "Japan",
            "Greece",
            "South",
            "China",
            "Cuba",
            "Iran",
            "Honduras",
            "Philippines",
            "Italy",
            "Poland",
            "Jamaica",
            "Vietnam",
            "Mexico",
            "Portugal",
            "Ireland",
            "France",
            "Dominican-Republic",
            "Laos",
            "Ecuador",
            "Taiwan",
            "Haiti",
            "Columbia",
            "Hungary",
            "Guatemala",
            "Nicaragua",
            "Scotland",
            "Thailand",
            "Yugoslavia",
            "El-Salvador",
            "Trinadad&Tobago",
            "Peru",
            "Hong",
            "Holand-Netherlands",
        ),
    }

    def __init__(
        self,
        root: Union[str, os.PathLike],
        train: bool = True,
        raw: bool = False,
        download: bool = False,
        output_fn: Optional[Callable[[str], None]] = print,
    ):
        """
        Loads the `Adult <https://archive.ics.uci.edu/dataset/2/adult>`_ dataset.

        :param root: The root directory where the Adult folder is placed or
          is to be downloaded to if download is set to True.
        :param train: Whether to retrieve the training set or test set of
          the dataset.
        :param raw: When :code:`True`, no one-hot encoding and standardization
         is applied to the downloaded data.
         Rows with missing values are dropped in any case.
        :param download: Whether to download the Adult dataset from
          https://archive.ics.uci.edu/dataset/2/adult if it is not
          present in the root directory.
        :param output_fn: A function for producing command line output.
          For example, :code:`print` or :code:`logging.info`.
          Pass `None` to turn off command line output.
        """
        self._train = train
        super().__init__(root, raw, download, output_fn)

    def _raw_files(self) -> tuple[str, ...]:
        return tuple(self.checksums)

    def _data_files(self) -> tuple[str, ...]:
        if not self._raw:
            return ("train.csv", "test.csv")
        else:
            return ("train_raw.csv", "test_raw.csv")

    def _data_file(self) -> str:
        file = "train" if self._train else "test"
        file += "_raw" if self._raw else ""
        return f"{file}.csv"

    def _target_column(self) -> str:
        return "income"

    def _download(self):
        self._download_zip(self.dataset_url, self.checksums)

    def _load_raw(self) -> Tuple[pandas.DataFrame, pandas.DataFrame]:
        all_columns = list(self.variables.keys()) + [self._target_column()]
        train_data: pandas.DataFrame = pandas.read_csv(
            self.files_dir / self.files_to_download["train"],
            header=None,
            index_col=False,
            names=all_columns,
        )
        test_data: pandas.DataFrame = pandas.read_csv(
            self.files_dir / self.files_to_download["test"],
            header=0,  # first row contains a note that we throw away
            index_col=False,
            names=all_columns,
        )
        return train_data, test_data

    _label_map = {"<=50K": False, ">50K": True, "<=50K.": False, ">50K.": True}

    def _preprocess(
        self, train_raw: pandas.DataFrame, test_raw: pandas.DataFrame
    ) -> Tuple[pandas.DataFrame, pandas.DataFrame]:
        self._output("Preprocessing data...")
        # preprocessing closely follows:
        # https://github.com/eth-sri/lcifr/blob/master/code/datasets/adult.py
        train_data, test_data = self._strip_strings(train_raw, test_raw)

        # transform the dataset: drop rows with missing values
        # missing values are encoded as ? in the original tables
        train_data, test_data = self._remove_missing_values(train_data, test_data)

        # map labels to (uniform) boolean values
        train_data.replace(self._label_map, inplace=True)
        test_data.replace(self._label_map, inplace=True)

        train_data, test_data = super()._preprocess(train_data, test_data)

        self._output("Preprocessing finished.")
        return train_data, test_data

    def _variables(self) -> Dict[str, Optional[Tuple[str, ...]]]:
        return self.variables



# Copyright (c) 2023 David Boetius
# Licensed under the MIT license
import logging

from fairnessdatasets import Adult


def test_print(tmp_path):
    Adult(root=tmp_path, train=False, download=True, output_fn=print)


def test_logging_info(tmp_path, caplog):
    caplog.set_level(logging.INFO)
    Adult(root=tmp_path, train=False, download=True, output_fn=logging.info)
    assert len(caplog.records) > 0


def test_no_output(tmp_path, caplog):
    caplog.set_level(logging.INFO)
    Adult(root=tmp_path, train=False, download=True, output_fn=None)
    assert len(caplog.records) == 0



# Copyright (c) 2024 David Boetius
# Licensed under the MIT license
import pytest

from fairnessdatasets import Adult, SouthGerman, Default, LawSchool


@pytest.mark.parametrize(
    "dataset_class,init_kwargs,dataset_path,expected_data_shape,expected_targets_shape",
    [
        (Adult, {}, "adult_path", (30162, 104), (30162,)),
        (Adult, {"train": False}, "adult_path", (15060, 104), (15060,)),
        (Adult, {"raw": True}, "adult_raw_path", (30162, 14), (30162,)),
        (Adult, {"train": False, "raw": True}, "adult_raw_path", (15060, 14), (15060,)),
        (Default, {}, "default_path", (30000, 95), (30000,)),
        (Default, {"raw": True}, "default_raw_path", (30000, 23), (30000,)),
        (LawSchool, {}, "law_school_path", (20888, 10), (20888,)),
        (LawSchool, {"raw": True}, "law_school_raw_path", (20888, 5), (20888,)),
        (SouthGerman, {"raw": False}, "south_german_path", (1000, 72), (1000,)),
        (SouthGerman, {"raw": True}, "south_german_raw_path", (1000, 20), (1000,)),
    ],
    ids=[
        "Adult-train",
        "Adult-test",
        "Adult-raw-train",
        "Adult-raw-test",
        "Default",
        "Default-raw",
        "LawSchool",
        "LawSchool-raw",
        "SouthGerman",
        "SouthGerman-raw",
    ],
)
def test_data_shapes(
    dataset_class,
    init_kwargs,
    dataset_path,
    expected_data_shape,
    expected_targets_shape,
    request,
):
    dataset_path = request.getfixturevalue(dataset_path)
    dataset = dataset_class(root=dataset_path, **init_kwargs)
    assert dataset.data.shape == expected_data_shape
    assert dataset.targets.shape == expected_targets_shape



"""
PyTorch dataset wrapper for the
`Adult (Census Income) <https://archive.ics.uci.edu/dataset/2/adult>`_ dataset.
"""
__version__ = "0.1.0"

from .adult import Adult
from .default import Default
from .law_school import LawSchool
from .south_german import SouthGerman



# Copyright (c) 2024 David Boetius
# Licensed under the MIT license
import pytest

from fairnessdatasets import Adult, SouthGerman, Default, LawSchool


@pytest.mark.parametrize(
    "dataset_class,init_kwargs",
    [
        (Adult, {}),
        (Adult, {"train": False}),
        (Adult, {"raw": True}),
        (Adult, {"train": False, "raw": True}),
        (Default, {}),
        (Default, {"raw": True}),
        (LawSchool, {}),
        (LawSchool, {"raw": True}),
        (SouthGerman, {"raw": False}),
        (SouthGerman, {"raw": True}),
    ],
    ids=[
        "Adult-train",
        "Adult-test",
        "Adult-raw-train",
        "Adult-raw-test",
        "Default",
        "Default-raw",
        "LawSchool",
        "LawSchool-raw",
        "SouthGerman",
        "SouthGerman-raw",
    ],
)
def test_download(dataset_class, init_kwargs, tmp_path):
    dataset = dataset_class(root=tmp_path, download=True, **init_kwargs)
    assert len(dataset[0]) == 2


@pytest.mark.parametrize(
    "dataset_class",
    [Adult, SouthGerman, Default],
)
def test_download_alongside_raw(dataset_class, tmp_path):
    raw = dataset_class(root=tmp_path, raw=True, download=True)
    assert len(raw[0]) == 2
    standard = dataset_class(root=tmp_path, raw=False, download=True)
    assert len(standard[0]) == 2

    assert len(raw.columns) < len(standard.columns)


@pytest.mark.parametrize("raw", [False, True], ids=["default", "raw"])
def test_download_south_german(raw: bool, tmp_path):
    dataset = SouthGerman(root=tmp_path, raw=raw, download=True)
    assert len(dataset[0]) == 2


def test_download_string_dir(tmp_path):
    dataset = SouthGerman(root=str(tmp_path), download=True)
    assert len(dataset[0]) == 2



# Copyright (c) 2024 David Boetius
# Licensed under the MIT license
import pytest

from fairnessdatasets import Adult, Default, LawSchool, SouthGerman


@pytest.mark.parametrize(
    "variable,expected_num_columns",
    [
        ("sex", 2),
        ("race", 5),
        ("relationship", 6),
        ("native-country", 41),
        ("age", 1),
    ],
)
def test_column_indices_adult(variable, expected_num_columns, adult_path):
    dataset = Adult(adult_path, train=False)
    assert len(dataset.column_indices(variable)) == expected_num_columns


@pytest.mark.parametrize(
    "variable",
    [
        "sex",
        "race",
        "relationship",
        "native-country",
        "age",
    ],
)
def test_sensitive_attribute_adult_raw(variable, adult_raw_path):
    dataset = Adult(adult_raw_path, train=False, raw=True)
    assert len(dataset.column_indices(variable)) == 1


@pytest.mark.parametrize(
    "dataset_class,init_kwargs,dataset_path,variable,expected_num_columns",
    [
        (Adult, {}, "adult_path", "sex", 2),
        (Adult, {}, "adult_path", "race", 5),
        (Adult, {}, "adult_path", "relationship", 6),
        (Adult, {}, "adult_path", "native-country", 41),
        (Adult, {}, "adult_path", "age", 1),
        (Adult, {}, "adult_path", "hours-per-week", 1),
        (Adult, {"raw": True}, "adult_raw_path", "sex", 1),
        (Adult, {"raw": True}, "adult_raw_path", "race", 1),
        (Adult, {"raw": True}, "adult_raw_path", "relationship", 1),
        (Adult, {"raw": True}, "adult_raw_path", "native-country", 1),
        (Adult, {"raw": True}, "adult_raw_path", "age", 1),
        (Adult, {"raw": True}, "adult_raw_path", "hours-per-week", 1),
        (Default, {}, "default_path", "SEX", 2),
        (Default, {}, "default_path", "EDUCATION", 4),
        (Default, {}, "default_path", "MARRIAGE", 3),
        (Default, {}, "default_path", "PAY_0", 12),
        (Default, {}, "default_path", "PAY_3", 12),
        (Default, {}, "default_path", "LIMIT_BAL", 1),
        (Default, {}, "default_path", "PAY_AMT5", 1),
        (Default, {"raw": True}, "default_raw_path", "SEX", 1),
        (Default, {"raw": True}, "default_raw_path", "EDUCATION", 1),
        (Default, {"raw": True}, "default_raw_path", "MARRIAGE", 1),
        (Default, {"raw": True}, "default_raw_path", "PAY_0", 1),
        (Default, {"raw": True}, "default_raw_path", "PAY_3", 1),
        (Default, {"raw": True}, "default_raw_path", "LIMIT_BAL", 1),
        (Default, {"raw": True}, "default_raw_path", "PAY_AMT5", 1),
        (LawSchool, {}, "law_school_path", "race1", 5),
        (LawSchool, {}, "law_school_path", "gender", 2),
        (LawSchool, {}, "law_school_path", "lsat", 1),
        (LawSchool, {}, "law_school_path", "zgpa", 1),
        (LawSchool, {"raw": True}, "law_school_raw_path", "race1", 1),
        (LawSchool, {"raw": True}, "law_school_raw_path", "gender", 1),
        (LawSchool, {"raw": True}, "law_school_raw_path", "lsat", 1),
        (LawSchool, {"raw": True}, "law_school_raw_path", "zgpa", 1),
        (SouthGerman, {}, "south_german_path", "credit_history", 5),
        (SouthGerman, {}, "south_german_path", "personal_status_sex", 4),
        (SouthGerman, {}, "south_german_path", "foreign_worker", 2),
        (SouthGerman, {}, "south_german_path", "age", 1),
        (SouthGerman, {"raw": True}, "south_german_raw_path", "credit_history", 1),
        (SouthGerman, {"raw": True}, "south_german_raw_path", "personal_status_sex", 1),
        (SouthGerman, {"raw": True}, "south_german_raw_path", "foreign_worker", 1),
        (SouthGerman, {"raw": True}, "south_german_raw_path", "age", 1),
    ],
)
def test_column_indices_default(
    dataset_class,
    init_kwargs,
    dataset_path,
    variable,
    expected_num_columns,
    request,
):
    dataset_path = request.getfixturevalue(dataset_path)
    dataset = dataset_class(root=dataset_path, **init_kwargs)
    assert len(dataset.column_indices(variable)) == expected_num_columns



# Copyright (c) 2024 David Boetius
# Licensed under the MIT license
import itertools
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Callable, Sequence, Tuple, Union

import os
from pathlib import Path

import numpy as np
import requests
import hashlib
from zipfile import ZipFile

import pandas
import torch
from torch.utils.data import Dataset


class CSVDataset(Dataset, ABC):
    """
    Base class for fairness datasets.
    Provides methods for downloading and processing data.

    Attributes:
    - `files_dir`: Where the data files are stored or downloaded to.
      Value: Root path (user specified) / `type(self).__name__)`
    - `data`: The dataset features (x values)
    - `targets` The dataset targets (y values)
    """

    def __init__(
        self,
        root: Union[str, os.PathLike],
        download: bool = False,
        output_fn: Optional[Callable[[str], None]] = print,
    ):
        """
        Creates a new :code:`FairnessDataset`.

        :param root: The root directory containing the data.
         If :code:`download=True` and the data isn't present in :code:`root`,
         it is downloaded to :code:`root`.
        :param download: Whether to download the dataset if it is not
          present in the :code:`root` directory.
        :param output_fn: A function for producing command line output.
          For example, :code:`print` or :code:`logging.info`.
          Pass `None` to turn off command line output.

        """
        if output_fn is None:

            def do_nothing(_):
                pass

            self.__output_fn = do_nothing
        else:
            self.__output_fn = output_fn

        self.files_dir = Path(root, type(self).__name__)
        if any(not (self.files_dir / file).exists() for file in self._raw_files()):
            if not download:
                raise RuntimeError(
                    "Dataset not found. Download it by passing download=True."
                )
            os.makedirs(self.files_dir, exist_ok=True)
            self._download()

        if any(not (self.files_dir / file).exists() for file in self._data_files()):
            data = self._load_raw()
            data = self._preprocess(*data)
            self._store_preprocessed(*data)

        table = self._load_preprocessed(self._data_file())
        data = table.drop(self._target_column(), axis=1)
        targets = table[self._target_column()]

        self.data = torch.tensor(
            data.values.astype(np.float64), dtype=torch.get_default_dtype()
        )
        self.targets = torch.tensor(targets.values.astype(np.int64))

    def _raw_files(self) -> tuple[str, ...]:
        """
        The downloaded files before preprocessing.

        Used for checking whether the dataset was already downloaded.
        """
        return ("raw.csv",)

    def _data_files(self) -> tuple[str, ...]:
        """
        The preprocessed data files.

        Used for checking whether the dataset is available in preprocessed form.
        """
        return ("data.csv",)

    def _data_file(self) -> str:
        """
        The data file containing the data of this instance
        (e.g. training data file vs test data file).

        Returns "data.csv" by default.
        """
        return "data.csv"

    def _target_column(self) -> str:
        """
        The column in the data csv containing the target variable.

        Default value: `"target"`
        """
        return "target"

    @abstractmethod
    def _download(self):
        """
        Download and extract the dataset to :code:`self.files_dir` so that
        the raw data can be loaded using :code:`self._load_raw()`.
        """
        raise NotImplementedError()

    def _load_raw(self) -> Tuple[pandas.DataFrame, ...]:
        """
        Load the raw data (before preprocessing) from :code:`self.files_dir`.
        By default, loads the files in :code:`self._raw_files` as csv files.

        :return: The raw data as :code:`pandas.DataFrames`.
        """
        return tuple(
            pandas.read_csv(self.files_dir / file) for file in self._raw_files()
        )

    @abstractmethod
    def _preprocess(self, *data: pandas.DataFrame) -> Tuple[pandas.DataFrame, ...]:
        """
        Preprocess downloaded data.

        :param data: The downloaded data.
        :return: The preprocessed data as :code:`pandas.DataFrames`.
        """
        raise NotImplementedError()

    def _store_preprocessed(self, *data: pandas.DataFrame):
        """
        Stores preprocessed data.

        The default implementation assumes that the order of :code:`data` matches
        the order of :code:`self._data_files` and stores the data in the corresponding
        files in the CSV format.
        Overwrite if this does not match your needs.
        """
        for file, table in zip(self._data_files(), data):
            table.to_csv(self.files_dir / file, index=False)

    def _load_preprocessed(self, file):
        """
        Loads preprocessed data.

        The default implementation loads :code:`file` as a csv.

        :param file: The file name of the data to load.
        """
        return pandas.read_csv(Path(self.files_dir, file))

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, torch.Tensor]:
        return self.data[index], self.targets[index]

    def __len__(self):
        return len(self.targets)

    def _download_zip(
        self,
        dataset_url: str,
        file_checksums: Dict[str, str],
    ):
        """
        Download and extract dataset .zip files.
        """
        self._output(f"Downloading {type(self).__name__} data...")
        dataset_path = self.files_dir / "dataset.zip"
        try:
            dataset_path.touch(exist_ok=False)
            result = requests.get(dataset_url, stream=True)
            with open(dataset_path, "wb") as dataset_file:
                for chunk in result.iter_content(chunk_size=256):
                    dataset_file.write(chunk)
            with ZipFile(dataset_path) as dataset_archive:
                for file_name in file_checksums:
                    dataset_archive.extract(file_name, self.files_dir)
        finally:
            dataset_path.unlink(missing_ok=True)

        self._output("Checking integrity of downloaded files...")
        for file_name, checksum in file_checksums.items():
            file_path = self.files_dir / file_name
            downloaded_file_checksum = self._sha256sum(file_path)
            if checksum != downloaded_file_checksum:
                raise RuntimeError(
                    f"Downloaded file has different checksum than expected: {file_name}. "
                    f"Expected sha256 checksum: {checksum}"
                )
        self._output("Download finished.")

    def _download_file(self, file_url: str, file_name: str, file_checksum: str):
        """
        Download a single file.

        :param file_name: The name of the file in which to store the download file.
        """
        self._output(f"Downloading {type(self).__name__} data...")
        file_path = self.files_dir / file_name
        file_path.touch(exist_ok=False)
        result = requests.get(file_url, stream=True)
        with open(file_path, "wb") as out_file:
            for chunk in result.iter_content(chunk_size=256):
                out_file.write(chunk)

        self._output("Checking integrity of downloaded files...")
        downloaded_file_checksum = self._sha256sum(file_path)
        if file_checksum != downloaded_file_checksum:
            raise RuntimeError(
                f"Downloaded file has different checksum than expected: {file_name}. "
                f"Expected sha256 checksum: {file_checksum}"
            )
        self._output("Download finished.")

    @staticmethod
    def _strip_strings(*data: pandas.DataFrame) -> Tuple[pandas.DataFrame, ...]:
        """Strips all strings in several :code:`DataFrames`."""
        return tuple(
            table.map(lambda val: val.strip() if isinstance(val, str) else val)
            for table in data
        )

    @staticmethod
    def _remove_missing_values(
        *data: pandas.DataFrame, marker="?"
    ) -> Tuple[pandas.DataFrame, ...]:
        """
        Removes rows with missing values from table.
        Modifies the data in-place.

        :param marker: The value marking missing values.
        :return: The preprocessed data (same as :code:`*data`).
        """
        for table in data:
            table.replace(to_replace=marker, value=np.nan, inplace=True)
            table.dropna(axis=0, inplace=True)
        return data

    @staticmethod
    def _categorical_to_integer(
        *data: pandas.DataFrame, variables: Sequence[str]
    ) -> Tuple[pandas.DataFrame, ...]:
        """
        Replaces string values of categorical attributes with integers.
        Modifies the data in-place.

        :param data: The tables to preprocess.
        :param variables: The categorical variables of the data as a mapping from
         variable names to variable values.
        :return: The preprocessed data (same as :code:`*data`).
        """
        for variable, values in variables.items():
            remapping = {value: index for index, value in enumerate(values)}
            for table in data:
                table.replace(remapping, inplace=True)
        return data

    @staticmethod
    def _encode_one_hot(
        *data: pandas.DataFrame, variables: Dict[str, Tuple[str, ...]], columns=None
    ) -> Tuple[pandas.DataFrame, ...]:
        """
        Applies a one-hot encoding to categorical variables.

        :param data: The tables to preprocess.
        :param variables: The categorical variables of the data as a mapping from
         variable names to variable values.
        :param columns: An optional set of columns that the tables should posses
         after preprocessing.
         Allows for adding columns for values that do not appear in the dataset.
        :return: The preprocessed data.
        """
        tables = data

        # one-hot encode all categorical variables
        tables = tuple(
            pandas.get_dummies(table, columns=variables.keys(), prefix_sep="=")
            for table in tables
        )

        # some tables may not contain all values of a categorical variable
        # make sure all tables have the same columns
        if columns is None:
            columns = set()
            for table in tables:
                columns.update(table.columns)
        for col in columns:
            for table in tables:
                if col not in table.columns:
                    table.insert(loc=0, column=col, value=0.0)
        return tables

    @staticmethod
    def _standardize(
        *data: pandas.DataFrame,
        variables: Sequence[str],
        reference_data: pandas.DataFrame,
    ) -> Tuple[pandas.DataFrame, ...]:
        """
        Z-score normalizes (standardizes) continuous variables.
        Modifies the data in-place.

        :param data: The tables to preprocess.
        :param variables: The continuous variables
        :param reference_data: The data to use for computing means and
         standard deviations of the continuous variables.
        :return: The preprocessed data (same as :code:`*data`).
        """
        # standardise continuous columns (z score)
        for col in variables:
            mean = reference_data[col].mean()
            std = reference_data[col].std()
            for table in data:
                table[col] = (table[col] - mean) / std
        return data

    @staticmethod
    def _sha256sum(path):
        # based on: https://stackoverflow.com/a/3431838/10550998 by quantumSoup
        # License: CC-BY-SA
        hash_ = hashlib.sha256()
        with open(path, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                hash_.update(chunk)
        return hash_.hexdigest()

    def _output(self, message: str):
        """Logging utility."""
        self.__output_fn(message)


class DefaultPreprocessing(CSVDataset, ABC):
    """
    A dataset that applies a one-hot encoding to all categorical variables
    and standardizes (z-score normalizes) the continuous variables.
    The class allows turning this normalization of by passing :code:`raw=True` to
    the initializer.

    Provides the :code:`columns` attribute and
    """

    def __init__(
        self,
        root: Union[str, os.PathLike],
        raw: bool = False,
        download: bool = False,
        output_fn: Optional[Callable[[str], None]] = print,
    ):
        """
        Creates a new :code:`NormalizedDataset`.
        Setting :code:`raw=True` turns off normalization.
        """
        self._raw = raw

        if not self._raw:
            self.columns = tuple(
                itertools.chain(
                    *(
                        [col_name]
                        if values is None
                        else [f"{col_name}={value}" for value in values]
                        for col_name, values in self._variables().items()
                    )
                )
            )
        else:
            self.columns = tuple(col_name for col_name in self._variables())

        super().__init__(root, download, output_fn)

    def column_indices(self, variable: str) -> Tuple[int, ...]:
        assert variable in self._variables()
        return tuple(
            i for i, col in enumerate(self.columns) if col.startswith(variable)
        )

    @abstractmethod
    def _variables(self) -> Dict[str, Optional[Tuple[str, ...]]]:
        """
        The variables and variable values of the dataset.
        Return a mapping from variable names to variable values.
        For non-categorical variables, the variable value is :code:`None`.
        For categorical variables, the variable value is a tuple of strings.
        """
        raise NotImplementedError()

    def _data_files(self) -> tuple[str, ...]:
        if not self._raw:
            return ("data.csv",)
        else:
            return ("data_raw.csv",)

    def _data_file(self) -> str:
        if not self._raw:
            return "data.csv"
        else:
            return "data_raw.csv"

    def _preprocess(self, *data: pandas.DataFrame) -> Tuple[pandas.DataFrame, ...]:
        categorical = {
            var: tuple(vals)
            for var, vals in self._variables().items()
            if vals is not None
        }
        continuous = [var for var, vals in self._variables().items() if vals is None]
        if not self._raw:
            data = self._encode_one_hot(
                *data, variables=categorical, columns=self.columns
            )
            data = self._standardize(
                *data, variables=continuous, reference_data=self._reference_data(*data)
            )
        else:
            data = self._categorical_to_integer(*data, variables=categorical)

        # reorder dataset columns and rename
        all_columns = list(self.columns) + [self._target_column()]
        return tuple(table[all_columns] for table in data)

    def _reference_data(self, *data):
        """
        Select the dataset to use for obtaining mean and standard deviation
        estimates for standardization.
        """
        return data[0]



# Copyright (c) 2024 Eric Tepper, David Boetius
# Licensed under the MIT license
from typing import Callable, Dict, Optional, Tuple, Union

import os

import pandas

from .base import DefaultPreprocessing


class Default(DefaultPreprocessing):
    """
    The `Default of credit card clients <https://archive.ics.uci.edu/dataset/350/default+of+credit+card+clients>`_
    dataset (Default for short).

    The dataset is preprocessed by:
     - Remove the "ID" column.
     - one-hot encoding all categorical attributes
     - applying z-score normalization to all continuous variables
    The last two preprocessing steps are optional and can be turned off by
    passing :code:`raw=True` to the initializer.

    Class Attributes:
    - `dataset_url`: The URL the Default dataset is downloaded from.
    - `file_to_download`: The file to download from `dataset_url`.
    - `checksums`: The checksums of the files to download from `dataset_url`.
    - `variables`: The variables of the Default dataset,
       together with the values they may take on and the interpretation of the values.

    Attributes:
    - `columns`: Column labels for the tensors in this dataset
      (after one-hot encoding, if applied).
    - `files_dir`: Where the data files are stored or downloaded to.
      Value: Root path (user specified) / `type(self).__name__)`
    - `data`: The dataset features (x values)
    - `targets` The dataset targets (y values)
    """

    dataset_url = "https://archive.ics.uci.edu/static/public/350/default+of+credit+card+clients.zip"
    file_to_download = "default of credit card clients.xls"
    checksums = {
        "default of credit card clients.xls": "30c6be3abd8dcfd3e6096c828bad8c2f011238620f5369220bd60cfc82700933",
    }

    PAYMENT_VALUES = {
        -2: "-2",
        -1: "pay duly",
        0: "0",
        1: "payment delay for one month",
        2: "payment delay for two months",
        3: "payment delay for three months",
        4: "payment delay for four months",
        5: "payment delay for five months",
        6: "payment delay for six months",
        7: "payment delay for seven months",
        8: "payment delay for eight months",
        9: "payment delay for nine months and above",
    }
    variables = {
        "LIMIT_BAL": None,
        "SEX": {
            1: "male",
            2: "female",
        },
        "EDUCATION": {
            1: "graduate school",
            2: "university",
            3: "high school",
            4: "others",
        },
        "MARRIAGE": {
            1: "married",
            2: "single",
            3: "others",
        },
        "AGE": None,
        "PAY_0": PAYMENT_VALUES,
        "PAY_2": PAYMENT_VALUES,
        "PAY_3": PAYMENT_VALUES,
        "PAY_4": PAYMENT_VALUES,
        "PAY_5": PAYMENT_VALUES,
        "PAY_6": PAYMENT_VALUES,
        "BILL_AMT1": None,
        "BILL_AMT2": None,
        "BILL_AMT3": None,
        "BILL_AMT4": None,
        "BILL_AMT5": None,
        "BILL_AMT6": None,
        "PAY_AMT1": None,
        "PAY_AMT2": None,
        "PAY_AMT3": None,
        "PAY_AMT4": None,
        "PAY_AMT5": None,
        "PAY_AMT6": None,
    }
    _variables_no_desc = {
        variable: None if values is None else tuple(values)
        for variable, values in variables.items()
    }

    def __init__(
        self,
        root: Union[str, os.PathLike],
        raw: bool = False,
        download: bool = False,
        output_fn: Optional[Callable[[str], None]] = print,
    ):
        """
        Loads the `Default <https://archive.ics.uci.edu/dataset/350/default+of+credit+card+clients>`_
        dataset.

        :param root: The root directory where the Default folder is placed or
          is to be downloaded to if download is set to True.
        :param raw: When :code:`True`, no one-hot encoding and standardization
         is applied to the downloaded data.
        :param download: Whether to download the Default of credit card clients dataset from
          https://archive.ics.uci.edu/dataset/350/default+of+credit+card+clients if it is not
          present in the root directory.
        :param output_fn: A function for producing command line output.
          For example, :code:`print` or :code:`logging.info`.
          Pass `None` to turn off command line output.
        """
        super().__init__(root, raw, download, output_fn)

    def _raw_files(self) -> tuple[str, ...]:
        return (self.file_to_download,)

    def _target_column(self) -> str:
        return "default payment next month"

    def _download(self):
        self._download_zip(self.dataset_url, self.checksums)

    def _load_raw(self) -> Tuple[pandas.DataFrame]:
        columns = list(self.variables.keys()) + [self._target_column()]
        data: pandas.DataFrame = pandas.read_excel(
            self.files_dir / self.file_to_download,
            sheet_name=0,
            header=None,
            names=columns,
            skiprows=2,
            index_col=0,
            dtype=int,
        )
        return (data,)

    def _preprocess(self, *data: pandas.DataFrame) -> Tuple[pandas.DataFrame]:
        self._output("Preprocessing data...")
        data = super()._preprocess(*data)
        self._output("Preprocessing finished.")
        return data

    def _variables(self) -> Dict[str, Optional[Tuple[str, ...]]]:
        return self._variables_no_desc



# Copyright (c) 2024 David Boetius
# Licensed under the MIT license
import pytest
from torch.utils.data import DataLoader

from fairnessdatasets import Adult, SouthGerman, Default


@pytest.fixture(
    scope="module",
    params=[
        (Adult, {"raw": False}, "adult_path"),
        (Adult, {"raw": True}, "adult_raw_path"),
        (SouthGerman, {"raw": False}, "south_german_path"),
        (SouthGerman, {"raw": True}, "south_german_raw_path"),
        (Default, {"raw": False}, "default_path"),
        (Default, {"raw": True}, "default_raw_path"),
    ],
    ids=[
        "Adult",
        "Adult-raw",
        "SouthGerman",
        "SouthGerman-raw",
        "Default",
        "Default-raw",
    ],
)
def dataset(request):
    dataset_class, init_kwargs, dataset_path = request.param
    dataset_path = request.getfixturevalue(dataset_path)
    return dataset_class(dataset_path, **init_kwargs)


def test_iterate(dataset):
    for i, (inputs, target) in enumerate(iter(dataset)):
        if i % 100 == 0:
            print(i, inputs, target)
        if i > 500:
            break


def test_data_loader(dataset):
    data_loader = DataLoader(dataset, batch_size=32, shuffle=True)
    for i, (inputs, targets) in enumerate(iter(data_loader)):
        if i % 100 == 0:
            print(i, inputs, targets)



# Copyright (c) 2024 David Boetius
# Licensed under the MIT license
from typing import Callable, Dict, Optional, Tuple, Union

import os

import pandas

from .base import DefaultPreprocessing


class LawSchool(DefaultPreprocessing):
    """
    The `Law School Admissions <https://eric.ed.gov/?id=ED469370>` dataset
    (Law School for short), downloaded from
    https://storage.googleapis.com/lawschool_dataset/bar_pass_prediction.csv.

    The dataset is preprocessed by:
     - Removing all columns, except for "race", "gender",
       "LSAT" (Law School Admission Test), "ZGPA" (Grade-Point Average),
       and "ZFYGPA" (First-Year Grade-Point Average).
       The "Z" presumably stands for these columns being z-score normalized.
     - removing rows (samples) with missing values
     - one-hot encoding all categorical attributes
     - applying z-score normalization to all continuous variables
    These preprocessing steps are optional and can be turned off by
    passing :code:`raw=True` to the initializer.

    Class Attributes:
    - `dataset_url`: The URL the Default dataset is downloaded from.
    - `checksum`: The checksum of the file to download from `dataset_url`.
    - `variables`: The selection of variables from the dataset that this class uses.
       Each variable is accompanied by the values which it can take on.
       For continuous variables, this entry is :code:`None`.

    Attributes:
    - `columns`: Column labels for the tensors in this dataset
      (after one-hot encoding, if applied).
    - `files_dir`: Where the data files are stored or downloaded to.
      Value: Root path (user specified) / `type(self).__name__)`
    - `data`: The dataset features (x values)
    - `targets` The dataset targets (y values)
    """

    dataset_url = (
        "https://storage.googleapis.com/lawschool_dataset/bar_pass_prediction.csv"
    )
    checksum = "c9a1d27b932bd697641c1c848d7e965232910285307e2edeb01c71edaeef11a2"

    variables = {
        "lsat": None,
        "zgpa": None,
        "zfygpa": None,
        "gender": ("female", "male"),
        "race1": ("black", "asian", "hisp", "white", "other"),
    }

    def __init__(
        self,
        root: Union[str, os.PathLike],
        raw: bool = False,
        download: bool = False,
        output_fn: Optional[Callable[[str], None]] = print,
    ):
        """
        Loads the `Law School <https://eric.ed.gov/?id=ED469370>`_ dataset.

        :param root: The root directory where the Default folder is placed or
          is to be downloaded to if download is set to True.
        :param raw: When :code:`True`, no one-hot encoding and standardization
         is applied to the downloaded data.
        :param download: Whether to download the Default of credit card clients dataset from
          https://storage.googleapis.com/lawschool_dataset/bar_pass_prediction.csv
          if it is not present in the root directory.
        :param output_fn: A function for producing command line output.
          For example, :code:`print` or :code:`logging.info`.
          Pass `None` to turn off command line output.
        """
        super().__init__(root, raw, download, output_fn)

    def _target_column(self) -> str:
        return "pass_bar"

    def _download(self):
        self._download_file(self.dataset_url, "raw.csv", self.checksum)

    def _load_raw(self) -> Tuple[pandas.DataFrame]:
        data: pandas.DataFrame = pandas.read_csv(
            self.files_dir / "raw.csv",
            header=0,
            index_col=0,
        )
        columns = list(self.variables.keys()) + [self._target_column()]
        return (data.filter(columns),)

    def _preprocess(self, *data: pandas.DataFrame) -> Tuple[pandas.DataFrame]:
        self._output("Preprocessing data...")
        data = self._strip_strings(*data)
        # remove rows with missing values
        data[0].dropna(axis=0, inplace=True)

        data = super()._preprocess(*data)
        self._output("Preprocessing finished.")
        return data

    def _variables(self) -> Dict[str, Optional[Tuple[str, ...]]]:
        return self.variables



# Copyright (c) 2024 Eric Tepper, David Boetius
# Licensed under the MIT license
from typing import Callable, Dict, Optional, Tuple, Union

import os

import pandas

from .base import DefaultPreprocessing


class SouthGerman(DefaultPreprocessing):
    """
    The `South German Credit <https://archive.ics.uci.edu/dataset/573/south+german+credit+update>`_ dataset.

    The dataset is preprocessed by:
     - one-hot encoding all categorical attributes
     - applying z-score normalization to all continuous variables
    These preprocessing steps are optional and can be turned off by
    passing :code:`raw=True` to the initializer.

    Class Attributes:
    - `dataset_url`: The URL the South German Credit dataset is downloaded from.
    - `file_to_download`: The file to download from `dataset_url`.
    - `checksums`: The checksums of the files to download from `dataset_url`.
    - `file`: The file containing the data after downloading.
    - `variables`: The variables of the South German Credit dataset,
       together with the values they may take on and the interpretation of the values.

    Attributes:
    - `columns`: Column labels for the tensors in this dataset
      (after one-hot encoding, if applied).
    - `files_dir`: Where the data files are stored or downloaded to.
      Value: Root path (user specified) / `type(self).__name__)`
    - `data`: The dataset features (x values)
    - `targets` The dataset targets (y values)
    """

    dataset_url = (
        "https://archive.ics.uci.edu/static/public/573/south+german+credit+update.zip"
    )
    file_to_download = "SouthGermanCredit.asc"
    checksums = {
        "SouthGermanCredit.asc": "5f363343f356ca38a0236baab849e472846399b2176ccc5bd686483dd8a7562f",
    }

    variables = {
        "status": {
            1: "no checking account",
            2: "... < 0 DM",
            3: "0<= ... < 200 DM",
            4: "... >= 200 DM",
        },
        "duration": None,
        "credit_history": {
            0: "delay in paying off in the past",
            1: "critical account/other credits elsewhere",
            2: "no credits taken/all credits paid back duly",
            3: "existing credits paid back duly till now",
            4: "all credits at this bank paid back duly",
        },
        "purpose": {
            0: "others",
            1: "car (new)",
            2: "car (used)",
            3: "furniture/equipment",
            4: "radio/television",
            5: "domestic appliances",
            6: "repairs",
            7: "education",
            8: "vacation",
            9: "retraining",
            10: "business",
        },
        "amount": None,
        "savings": {
            1: "unknown/no savings account",
            2: "... <  100 DM",
            3: "100 <= ... <  500 DM",
            4: "500 <= ... < 1000 DM",
            5: "... >= 1000 DM",
        },
        "employment_duration": {
            1: "unemployed",
            2: "< 1 yr",
            3: "1 <= ... < 4 yrs",
            4: "4 <= ... < 7 yrs",
            5: ">= 7 yrs",
        },
        "installment_rate": {
            1: ">= 35",
            2: "25 <= ... < 35",
            3: "20 <= ... < 25",
            4: "< 20",
        },
        "personal_status_sex": {
            1: "male : divorced/separated",
            2: "female : non-single or male : single",
            3: "male : married/widowed",
            4: "female : single",
        },
        "other_debtors": {
            1: "none",
            2: "co-applicant",
            3: "guarantor",
        },
        "present_residence": {
            1: "< 1 yr",
            2: "1 <= ... < 4 yrs",
            3: "4 <= ... < 7 yrs",
            4: ">= 7 yrs",
        },
        "property": {
            1: "unknown / no property",
            2: "car or other",
            3: "building soc. savings agr./life insurance",
            4: "real estate",
        },
        "age": None,
        "other_installment_plans": {
            1: "bank",
            2: "stores",
            3: "none",
        },
        "housing": {
            1: "for free",
            2: "rent",
            3: "own",
        },
        "number_credits": {
            1: "1",
            2: "2-3",
            3: "4-5",
            4: ">= 6",
        },
        "job": {
            1: "unemployed/unskilled - non-resident",
            2: "unskilled - resident",
            3: "skilled employee/official",
            4: "manager/self-empl./highly qualif. employee",
        },
        "people_liable": {
            1: "3 or more",
            2: "0 to 2",
        },
        "telephone": {
            1: "no",
            2: "yes (under customer name)",
        },
        "foreign_worker": {
            1: "yes",
            2: "no",
        },
    }
    _variables_no_desc = {
        variable: None if values is None else tuple(values)
        for variable, values in variables.items()
    }

    def __init__(
        self,
        root: Union[str, os.PathLike],
        raw: bool = False,
        download: bool = False,
        output_fn: Optional[Callable[[str], None]] = print,
    ):
        """
        Loads the `South German Credit <https://archive.ics.uci.edu/dataset/573/south+german+credit+update>`_ dataset.

        :param root: The root directory where the SouthGerman folder is placed or
          is to be downloaded to if download is set to True.
        :param raw: When :code:`True`, no one-hot encoding and standardization
         is applied to the downloaded data.
        :param download: Whether to download the South German Credit dataset from
          https://archive.ics.uci.edu/dataset/573/south+german+credit+update if it is not
          present in the root directory.
        :param output_fn: A function for producing command line output.
          For example, :code:`print` or :code:`logging.info`.
          Pass `None` to turn off command line output.
        """
        super().__init__(root, raw, download, output_fn)

    def _raw_files(self) -> tuple[str, ...]:
        return (self.file_to_download,)

    def _target_column(self) -> str:
        return "credit_risk"

    def _download(self):
        self._download_zip(self.dataset_url, self.checksums)

    def _load_raw(self) -> Tuple[pandas.DataFrame]:
        all_columns = list(self.variables.keys()) + [self._target_column()]
        data: pandas.DataFrame = pandas.read_csv(
            self.files_dir / self.file_to_download,
            index_col=False,
            names=all_columns,
            delimiter=" ",
            skiprows=1,  # skip first row since pandas does not recognize header row
            dtype=int,
        )
        return (data,)

    def _preprocess(self, *data: pandas.DataFrame) -> Tuple[pandas.DataFrame]:
        self._output("Preprocessing data...")
        data = super()._preprocess(*data)
        self._output("Preprocessing finished.")
        return data

    def _variables(self) -> Dict[str, Optional[Tuple[str, ...]]]:
        return self._variables_no_desc



import nox


@nox.session
@nox.parametrize(
    "python,torch_version,numpy_version,pandas_version,requests_version",
    [
        ("3.9", "1.9.0", "1.20.0", "2.1.0", "2.25.0"),
        ("3.10", "1.12.1", "1.22.0", "2.1.0", "2.27.0"),
        ("3.11", "2.1.2", "1.26.2", "2.1.4", "2.31.0"),
        ("3.12", "2.2.0", "1.26.3", "2.2.0", "2.31.0"),
    ],
    ids=["python3.9", "python3.10", "python3.11", "python3.12"],
)
def tests(session, torch_version, numpy_version, pandas_version, requests_version):
    session.install(f"torch=={torch_version}")
    session.install(f"numpy=={numpy_version}")
    session.install(f"pandas=={pandas_version}")
    session.install(f"requests=={requests_version}")
    session.install("pytest")
    session.install(".")
    session.run("pytest")
